/*
 * Copyright (c) 2012, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 *  ======== main_dsp.c ========
 *
 */

#include <stdlib.h>
#include <string.h>

/* xdctools header files */
#include <xdc/std.h>
#include <xdc/runtime/Diags.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/Log.h>
#include <xdc/runtime/System.h>

/* package header files */
#include <ti/ipc/Ipc.h>
#include <ti/ipc/MultiProc.h>
#include <ti/sysbios/BIOS.h>
#include <ti/sysbios/knl/Task.h>

/* local header files */
#include "Server.h"

/* private functions */
static Void smain(UArg arg0, UArg arg1);

/* private data */
static String           Mod_file = __FILE__;
static Int              Mod_line = 0;


/*
 *  ======== main ========
 */
Int main(Int argc, Char* argv[])
{
    Error_Block eb;
    Task_Params taskParams;

    Log_print0(Diags_INFO, "--> main:");

    /* check for debug flag */
    if ((argc >= 3) && (strcmp(argv[1], "-d") == 0)) {
        Int d_flag = atoi(argv[2]);
        if (d_flag & 0x1) {
            volatile Int spin = 1;
            while (spin);
        }
    }

    /* must initialize the error block before using it */
    Error_init(&eb);

    /* create main thread (interrupts not enabled in main on BIOS) */
    Task_Params_init(&taskParams);
    taskParams.instance->name = "smain";
    taskParams.arg0 = (UArg)argc;
    taskParams.arg1 = (UArg)argv;
    taskParams.priority = 8;
    taskParams.stackSize = 0x1000;
    Task_create(smain, &taskParams, &eb);

    if (Error_check(&eb)) {
        System_abort("main: failed to create application startup thread");
    }

    /* start scheduler, this never returns */
    BIOS_start();

    /* should never get here */
    Log_print0(Diags_INFO, "<-- main:");
    return (0);
}

/*
 *  ======== smain ========
 */
Void smain(UArg arg0, UArg arg1)
{
    Int         status = 0;
    Error_Block eb;
    UInt16      remoteProcId;
    Bool        running = TRUE;
    UInt        loop = 0;
    Int         argc = (Int)arg0;
    String *    argv = (String *)arg1;

    Log_print0(Diags_ENTRY | Diags_INFO, "--> smain:");

    Error_init(&eb);

    /* loop forever */
    while (running) {

        /* only one thread must call start */
        do {
            status = Ipc_start();

            if (status == Ipc_E_NOTREADY) {
                Task_sleep(1);
            }
        } while (status == Ipc_E_NOTREADY);

        if (status < 0) {
            Mod_line = __LINE__;
            goto leave;
        }

        /* attach to the remote processor */
        remoteProcId = MultiProc_getId("HOST");

        /* connect to remote processor */
        do {
            loop++;
            status = Ipc_attach(remoteProcId);

            if ((status == Ipc_E_NOTREADY) && ((loop & 0xFFF) == 0)) {
                Task_sleep(1);
            }
        } while (status == Ipc_E_NOTREADY);

        if (status < 0) {
            Mod_line = __LINE__;
            goto leave;
        }

        /* check for debug flag */
        if ((argc >= 3) && (strcmp(argv[1], "-d") == 0)) {
            Int d_flag = atoi(argv[2]);
            if (d_flag & 0x2) {
                volatile Int spin = 1;
                while (spin);
            }
        }

        /*
         *  BEGIN server phase
         */

        /* server setup phase */
        status = Server_setup();

        if (status < 0) {
            Mod_line = __LINE__;
            goto leave;
        }

        /* server execute phase */
        status = Server_run();

        if (status < 0) {
            Mod_line = __LINE__;
            goto leave;
        }

        /* server shutdown phase */
        status = Server_destroy();

        if (status < 0) {
            Mod_line = __LINE__;
            goto leave;
        }

        /*
         *  END server phase
         */

        /* disconnect from remote processor */
        do {
            status = Ipc_detach(remoteProcId);

            if ((status == Ipc_E_NOTREADY) && ((++loop & 0xFFF) == 0)) {
                Task_sleep(1);
            }
        } while (status == Ipc_E_NOTREADY);

        if (status < 0) {
            Mod_line = __LINE__;
            goto leave;
        }

        /* only one thread must call stop */
        Ipc_stop();

    } /* while (running) */

leave:
    /* report error */
    if (status < 0) {
        Log_error3("Error: smain: error=%d, file=%s, line=%d\n",
            (IArg)status, (IArg)Mod_file, (IArg)Mod_line);
    }

    Log_print1(Diags_EXIT, "<-- smain: %d", (IArg)status);
    return;
}
